from core.database import db

from schemas.tokens_schema import accessTokenCreate,refreshTokenCreate,accessTokenOut,refreshTokenOut
import asyncio
from datetime import datetime, timedelta
from dateutil import parser
from bson import ObjectId,errors
from fastapi import HTTPException

async def add_access_tokens(token_data:accessTokenCreate)->accessTokenOut:
    token = token_data.model_dump()
    token['role']="member"
    result = await db.accessToken.insert_one(token)
    tokn = await db.accessToken.find_one({"_id":result.inserted_id})
    accessToken = accessTokenOut(**tokn)
    
    return accessToken 
    

async def add_admin_access_tokens(token_data:accessTokenCreate)->accessTokenOut:
    token = token_data.model_dump()
    token['role']="admin"
    token['status']="inactive"
    result = await db.accessToken.insert_one(token)
    tokn = await db.accessToken.find_one({"_id":result.inserted_id})
    accessToken = accessTokenOut(**tokn)
    
    return accessToken 

async def update_admin_access_tokens(token:str)->accessTokenOut:
    updatedToken= await db.accessToken.find_one_and_update(filter={"_id":ObjectId(token)},update={"$set": {'status':'active'}},return_document=True)
    accessToken = accessTokenOut(**updatedToken)
    return accessToken
    
async def add_refresh_tokens(token_data:refreshTokenCreate)->refreshTokenOut:
    token = token_data.model_dump()
    result = await db.refreshToken.insert_one(token)
    tokn = await db.refreshToken.find_one({"_id":result.inserted_id})
    refreshToken = refreshTokenOut(**tokn)
    return refreshToken

async def delete_access_token(accessToken):
    # await db.refreshToken.delete_many({"previousAccessToken":accessToken})
    await db.accessToken.find_one_and_delete({'_id':ObjectId(accessToken)})
    
    
async def delete_refresh_token(refreshToken:str):
    try:
        obj_id=ObjectId(refreshToken)
    except errors.InvalidId:
        raise HTTPException(status_code=401,detail="Invalid Refresh Id")
    result = await db.refreshToken.find_one_and_delete({"_id":obj_id})
    if result:
        return True



def is_older_than_days(date_string, days=10):
    # Parse the ISO 8601 date string into a datetime object
    created_date = parser.isoparse(date_string)

    # Get the current time in UTC
    now = datetime.utcnow().replace(tzinfo=created_date.tzinfo)

    # Check if the difference is greater than the given number of days
    return (now - created_date) > timedelta(days=days)



async def get_access_tokens(accessToken:str):
    
    token = await db.accessToken.find_one({"_id": ObjectId(accessToken)})
    if token:
        if is_older_than_days(date_string=token['dateCreated'])==False:
            if token.get("role",None)=="member":
                tokn = accessTokenOut(**token)
                return tokn
            elif token.get("role",None)=="admin":
                if token.get('status',None)=="active":
                    tokn = accessTokenOut(**token)
                    return tokn
                else: 
                    return None
            else:
                return None
            
        else:
            delete_access_token(accessToken=str(token['_id'])) 
            return None
    else:
        print("No token found")
        return "None"
    
    
async def get_refresh_tokens(refreshToken:str):
    token = await db.refreshToken.find_one({"_id": ObjectId(refreshToken)})
    if token:
        tokn = refreshTokenOut(**token)
        return tokn

    else: return None
    
    
    
async def delete_all_tokens_with_user_id(userId:str):
    await db.refreshToken.delete_many(filter={"userId":userId})
    await db.accessToken.delete_many(filter={"userId":userId})


from bson import ObjectId
from pydantic import GetJsonSchemaHandler
from pydantic import BaseModel, EmailStr, Field,model_validator
from pydantic_core import core_schema
from datetime import datetime,timezone
from typing import Optional,List,Any




from string import Template
import os
from dotenv import load_dotenv

load_dotenv()

# Choose between 'sqlite' or 'mongodb'
DB_NAME = os.getenv("DB_NAME", "test").lower()


new_signin_warning_template_string=Template("""
<!DOCTYPE html>
<html lang="en">
  <head>
    <title>A New Login </title>
    <!--[if !mso]><!-- -->
    <meta http-equiv="X-UA-Compatible" content="IE=edge" />
    <!--<![endif]-->
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <style type="text/css">
      #outlook a { padding: 0; }
      .ReadMsgBody { width: 100%; }
      .ExternalClass { width: 100%; }
      .ExternalClass * { line-height: 100%; }
      body {
        margin: 0;
        padding: 0;
        -webkit-text-size-adjust: 100%;
        -ms-text-size-adjust: 100%;
      }
      table, td {
        border-collapse: collapse;
        mso-table-lspace: 0pt;
        mso-table-rspace: 0pt;
      }
      img {
        border: 0;
        height: auto;
        line-height: 100%;
        outline: none;
        text-decoration: none;
        -ms-interpolation-mode: bicubic;
      }
      p { display: block; margin: 13px 0; }
    </style>

    <!--[if !mso]><!-->
    <style type="text/css">
      @media only screen and (max-width: 480px) {
        @-ms-viewport { width: 320px; }
        @viewport { width: 320px; }
      }
    </style>
    <!--<![endif]-->

    <!--[if mso]>
    <xml>
      <o:OfficeDocumentSettings>
        <o:AllowPNG />
        <o:PixelsPerInch>96</o:PixelsPerInch>
      </o:OfficeDocumentSettings>
    </xml>
    <![endif]-->

    <!--[if lte mso 11]>
    <style type="text/css">
      .outlook-group-fix {
        width: 100% !important;
      }
    </style>
    <![endif]-->

    <style type="text/css">
      @media only screen and (min-width: 480px) {
        .mj-column-per-100 {
          width: 100% !important;
          max-width: 100%;
        }
      }
      @media only screen and (max-width: 480px) {
        table.full-width-mobile { width: 100% !important; }
        td.full-width-mobile { width: auto !important; }
      }

      h1 {
        font-family: -apple-system, system-ui, BlinkMacSystemFont;
        font-size: 24px;
        font-weight: 600;
        line-height: 24px;
        text-align: left;
        color: #333333;
        padding-bottom: 18px;
      }
      p {
        font-family: -apple-system, system-ui, BlinkMacSystemFont;
        font-size: 15px;
        font-weight: 300;
        line-height: 24px;
        text-align: left;
        color: #333333;
      }
      a {
        color: #0867ec;
        font-weight: 400;
      }
      a.footer-link {
        color: #888888;
      }
      strong {
        font-weight: 500;
      }
    </style>
  </head>

  <body style="background-color: #ffffff">
    <div style="display: none; font-size: 1px; color: #ffffff; line-height: 1px; max-height: 0px; max-width: 0px; opacity: 0; overflow: hidden;">
      Your $DB_NAME account has been accessed from a new IP address
    </div>

    <div style="background-color: #ffffff">
      <!--[if mso | IE]>
      <table align="center" border="0" cellpadding="0" cellspacing="0" style="width:600px;" width="600">
        <tr><td style="line-height:0px;font-size:0px;mso-line-height-rule:exactly;">
      <![endif]-->

      <div style="background: #ffffff; background-color: #ffffff; margin: 0px auto; max-width: 600px;">
        <table align="center" border="0" cellpadding="0" cellspacing="0" role="presentation" style="background: #ffffff; background-color: #ffffff; width: 100%">
          <tbody>
            <tr>
              <td style="direction: ltr; font-size: 0px; padding: 20px 0; text-align: center; vertical-align: top;">
                <!--[if mso | IE]>
                <table role="presentation" border="0" cellpadding="0" cellspacing="0"><tr><td style="vertical-align:top;width:600px;">
                <![endif]-->

                <div class="mj-column-per-100 outlook-group-fix" style="font-size: 13px; text-align: left; direction: ltr; display: inline-block; vertical-align: top; width: 100%;">
                  <table border="0" cellpadding="0" cellspacing="0" role="presentation" style="vertical-align: top;" width="100%">
                    <tr>
                      <td align="left" style="font-size: 0px; padding: 10px 25px; word-break: break-word;">
                        <table border="0" cellpadding="0" cellspacing="0" role="presentation" style="border-collapse: collapse; border-spacing: 0px;">
                          <tbody>
                            <tr>
                              <td style="width: 54px; border-radius: 100%; ">
                                <img style=" width: 100%; height: 100%;  border-radius: 20%; transform-origin: center center; transform: scale(1.0); " alt="$DB_NAME logo"  height="auto" src="$helpful_img" style="border: 0; display: block; outline: none; text-decoration: none; height: auto; width: 100%;" width="24" />
                              </td>
                            </tr>
                          </tbody>
                        </table>
                      </td>
                    </tr>

                    <tr>
                      <td align="left" style="font-size: 0px; padding: 10px 25px 24px 25px; word-break: break-word;">
                        <div style="font-family: -apple-system, system-ui, BlinkMacSystemFont; font-size: 15px; font-weight: 300; line-height: 24px; text-align: left; color: #333333;">
                          <h1>We've noticed a new login</h1>
                          <p>Hi $firstName $lastName,</p>
                          <p>This is a routine security alert. Someone logged into your $DB_NAME account from a new IP address:</p>
                          <p>
                            <strong>Time:</strong> $time_data<br />
                            <strong>IP address:</strong> $ip_address<br />
                            <strong>Location:</strong> $location<br />
                            <strong>More Information:</strong> $extra_data
                          </p>
                          <p>If this was you, you can ignore this alert. If you noticed any suspicious activity on your account, please change your password on your email login and on your account page.</p>
                        </div>
                      </td>
                    </tr>

                    <tr>
                      <td align="left" style="font-size: 0px; padding: 10px 25px; word-break: break-word;">
                        <div style="font-family: -apple-system, system-ui, BlinkMacSystemFont; font-size: 15px; font-weight: 300; line-height: 24px; text-align: left; color: #333333;">
                          So long, and thanks for all the fish,<br />
                          <strong>The $DB_NAME Team</strong>
                        </div>
                      </td>
                    </tr>

                    <tr>
                      <td style="font-size: 0px; padding: 10px 25px; word-break: break-word;">
                        <p style="border-top: solid 1px #e8e8e8; font-size: 1; margin: 0px auto; width: 100%;"></p>
                        <!--[if mso | IE]>
                        <table align="center" border="0" cellpadding="0" cellspacing="0" style="border-top: solid 1px #e8e8e8; font-size: 1; margin: 0px auto; width: 550px;" role="presentation" width="550px">
                          <tr><td style="height: 0; line-height: 0">&nbsp;</td></tr>
                        </table>
                        <![endif]-->
                      </td>
                    </tr>

                    <tr>
                      <td align="left" style="font-size: 0px; padding: 10px 25px; word-break: break-word;">
                        <div style="font-family: -apple-system, system-ui, BlinkMacSystemFont; font-size: 12px; font-weight: 300; line-height: 24px; text-align: left; color: #888888;">
                          Somewhere Between Coffee & Code, Quiet Meadows, Earth 00000<br />
                          © 2025 $DB_NAME. LLC
                        </div>
                      </td>
                    </tr>

                    <tr>
                      <td align="left" style="font-size: 0px; padding: 10px 25px; word-break: break-word;">
                        <div style="font-family: -apple-system, system-ui, BlinkMacSystemFont; font-size: 12px; font-weight: 300; line-height: 24px; text-align: left; color: #888888;">
                          For questions contact <a href="mailto:support@x.ai" class="footer-link">support@$DB_NAME.com.ng</a>
                        </div>
                      </td>
                    </tr>
                  </table>
                </div>

                <!--[if mso | IE]></td></tr></table><![endif]-->
              </td>
            </tr>
          </tbody>
        </table>
      </div>
      <!--[if mso | IE]></td></tr></table><![endif]-->
    </div>
  </body>
</html>


""")

def generate_new_signin_warning_email_from_template(firstName,lastName,time_data,ip_address,location,extra_data):
    generated_email = new_signin_warning_template_string.safe_substitute(DB_NAME=DB_NAME,helpful_img="https://iili.io/3DKqndN.jpg",firstName=firstName,lastName=lastName,time_data=time_data,ip_address=ip_address,location=location,extra_data=extra_data )
    return generated_email


from pathlib import Path

def create_crud_file(name: str):
    db_name=name.lower()
    schema_path = Path.cwd() / "schemas" / f"{db_name}.py"
    repo_path = Path.cwd() / "repositories" / f"{db_name}.py"

    if not schema_path.exists():
        print(f"❌ Schema file {schema_path} not found.")
        return

    class_name = "".join([part.capitalize() for part in db_name.split("_")]) + "Base"
    update_class_name = "".join([part.capitalize() for part in db_name.split("_")]) + "Update"
    create_class_name = "".join([part.capitalize() for part in db_name.split("_")]) + "Create"
    out_class_name = "".join([part.capitalize() for part in db_name.split("_")]) + "Out"
    crud_code = f'''
from core.database import db
from schemas.{db_name} import {update_class_name},{create_class_name},{out_class_name}

def create_{db_name}({db_name}_data: {create_class_name})->{out_class_name}:
    {db_name}_dict = {db_name}_data.model_dump()
    result = db.{db_name}s.insert_one({db_name}_dict)
    returnable_result = {out_class_name}(**result)
    return returnable_result

def get_{db_name}(filter_dict: dict)->{out_class_name}:
    result = db.{db_name}s.find_one(filter_dict)
    returnable_result = {out_class_name}(**result)
    return returnable_result
    
def update_{db_name}(filter_dict: dict, {db_name}_data: {update_class_name})->{out_class_name}:
    result = db.{db_name}s.find_one_and_update(
    filter_dict,
    {"$set": {db_name}_data.model_dump()},
    return_document=ReturnDocument.AFTER
)
    returnable_result = {out_class_name}(**result)
    return returnable_result

def delete_{db_name}(filter_dict: dict):
    return db.{db_name}s.delete_one(filter_dict)
'''.strip()

    with open(repo_path, "w") as f:
        f.write(crud_code)

    print(f"✅ CRUD for '{db_name}' created in repository/{db_name}.py")


import click
from fasterapi.scaffolder.generate_project import create_project
from fasterapi.scaffolder.generate_crud import create_crud_file
from fasterapi.scaffolder.generate_schema import create_schema_file
from fasterapi.__version__ import __version__


@click.group()
@click.version_option(__version__, '--version', '-v', message='FasterAPI version %(version)s')
def cli():
    """⚡ FasterAPI CLI — Scaffold FastAPI apps with ease"""
    pass


@cli.command()
@click.argument("name")
def new(name):
    """Create a new FastAPI project."""
    create_project(name)

@cli.command()
@click.argument("name")
def make_crud(name):
    """Generate CRUD repository functions for a schema."""
    create_crud_file(name)
    
@cli.command()
@click.argument("name")
def make_schema(name):
    """Generate Pydantic classes templates for a schema."""
    create_schema_file(name)

if __name__ == "__main__":
    cli()



# auth.py
from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPBearer

from security.tokens import validate_admin_accesstoken,generate_refresh_tokens,generate_member_access_tokens, validate_member_accesstoken, validate_refreshToken,validate_member_accesstoken_without_expiration,generate_admin_access_tokens,validate_expired_admin_accesstoken
from security.encrypting_jwt import decode_jwt_token,decode_jwt_token_without_expiration
from repositories.tokens_repo import delete_access_token,update_admin_access_tokens
from schemas.tokens_schema import refreshedToken,accessTokenOut


token_auth_scheme = HTTPBearer()
async def verify_token(token: str = Depends(token_auth_scheme)):
    result = await validate_member_accesstoken(accessToken=token.credentials)
    
    if result==None:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid token"
        )
    else:
        decoded_access_token = await decode_jwt_token(token=token.credentials)
        return decoded_access_token
            
        
        
        
async def verify_token_and_refresh_token(token: str = Depends(token_auth_scheme)):
    decodedT= await decode_jwt_token_without_expiration(token=token.credentials)
    if decodedT['role']=="member":
        result = await validate_member_accesstoken_without_expiration(accessToken=token.credentials)
        
        accessTokenObj= await generate_member_access_tokens(userId=result.userId)
        
        refreshTokenObj = await generate_refresh_tokens(userId=result.userId,accessToken=accessTokenObj.accesstoken)
        await delete_access_token(result.accesstoken)
        refreshedTokens = refreshedToken(userId=result.userId,refreshToken=refreshTokenObj.refreshtoken,accessToken=accessTokenObj.accesstoken)
        if result==None:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid token"
            )
        else:
            return refreshedTokens
    elif decodedT['role']=="admin":
        result = await validate_expired_admin_accesstoken(accessToken=str(token.credentials))
        print("result",result)
        if result =="inactive":
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="You Can't Make Use of an Inactive AccessToken"
            )
        
        elif isinstance(result,accessTokenOut):
            accessTokenObj = await generate_admin_access_tokens(userId=result.userId)
            NewDecodedT = await decode_jwt_token(token=accessTokenObj.accesstoken)
            
            await update_admin_access_tokens(token=NewDecodedT['accessToken'])
            refreshTokenObj = await generate_refresh_tokens(userId=result.userId,accessToken=accessTokenObj.accesstoken)
            await delete_access_token(result.accesstoken)
            refreshedTokens = refreshedToken(userId=result.userId,refreshToken=refreshTokenObj.refreshtoken,accessToken=accessTokenObj.accesstoken)
            return refreshedTokens
        elif result==None:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid Token"
            )
           
async def verify_admin_token(token: str = Depends(token_auth_scheme)):
    try:
        result = await validate_admin_accesstoken(accessToken=str(token.credentials))

        if result==None:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid admin token"
            )
        elif result=="inactive":
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Admin Token hasn't been activated"
            )
        elif isinstance(result, accessTokenOut):
            
            decoded_access_token = await decode_jwt_token(token=token.credentials)
            return decoded_access_token
    except TypeError:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,detail="Access Token Expired")

async def verify_any_token(token:str=Depends(token_auth_scheme)):
    token_type = await decode_jwt_token(token=token.credentials)
    if isinstance(token_type,dict):
        if token_type['role']=='admin':
            return await verify_admin_token(token=token)
        elif token_type["role"]=='member':
            return await verify_token(token=token)
        else:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid token Type"
            )
    else:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid Token"
        )


import os
from dotenv import load_dotenv

load_dotenv()

# Choose between 'sqlite' or 'mongodb'
DB_TYPE = os.getenv("DB_TYPE", "sqlite").lower()

if DB_TYPE == "sqlite":
    # SQLite setup
    import sqlite3

    database_name = "db.db"

    class DBFunctions:
        def __init__(self, table_name):
            self.table_name = table_name

        @staticmethod
        def __insert(table_name: str, data: dict):
            if not table_name.isidentifier():
                raise ValueError("Invalid table name")

            keys = ", ".join(data.keys())
            placeholders = ", ".join("?" for _ in data)
            values = tuple(data.values())

            if table_name == "password_reset_token":
                query = f"INSERT OR REPLACE INTO {table_name} ({keys}) VALUES ({placeholders})"
            else:
                query = f"INSERT INTO {table_name} ({keys}) VALUES ({placeholders})"

            with sqlite3.connect(database_name) as conn:
                cursor = conn.cursor()
                cursor.execute(query, values)
                return cursor.lastrowid

        @staticmethod
        def __update(table_name: str, data: dict, filter_dict: dict):
            if not table_name.isidentifier():
                raise ValueError("Invalid table name")

            set_clause = ", ".join([f"{k} = ?" for k in data.keys()])
            where_clause = " AND ".join([f"{k} = ?" for k in filter_dict.keys()])
            values = list(data.values()) + list(filter_dict.values())

            query = f"UPDATE {table_name} SET {set_clause} WHERE {where_clause}"

            with sqlite3.connect(database_name) as conn:
                cursor = conn.cursor()
                cursor.execute(query, values)
                conn.commit()
                return cursor.rowcount

        @staticmethod
        def __delete(table_name: str, filter_dict: dict, limit: int = None):
            if not table_name.isidentifier():
                raise ValueError("Invalid table name")

            where_clause = " AND ".join([f"{k} = ?" for k in filter_dict])
            values = list(filter_dict.values())

            if limit is not None:
                query = f"""
                    DELETE FROM {table_name}
                    WHERE rowid IN (
                        SELECT rowid FROM {table_name}
                        WHERE {where_clause}
                        LIMIT ?
                    )
                """
                values.append(limit)
            else:
                query = f"DELETE FROM {table_name} WHERE {where_clause}"

            with sqlite3.connect(database_name) as conn:
                cursor = conn.cursor()
                cursor.execute(query, values)
                conn.commit()
                return cursor.rowcount

        def insert_one(self, data: dict) -> str:
            return self.__insert(table_name=self.table_name, data=data)

        def update_one(self, filter_dict: dict, data: dict) -> int:
            return self.__update(filter_dict=filter_dict, table_name=self.table_name, data=data)

        def delete_one(self, filter_dict: dict) -> int:
            return self.__delete(table_name=self.table_name, filter_dict=filter_dict, limit=1)

        def delete_many(self, filter_dict: dict, limit: int = None) -> int:
            return self.__delete(table_name=self.table_name, filter_dict=filter_dict, limit=limit)

        def find_one(self, filter_dict: dict) -> dict:
            if not filter_dict:
                raise ValueError("Filter dictionary cannot be empty.")

            with sqlite3.connect(database_name) as conn:
                conn.row_factory = sqlite3.Row
                cursor = conn.cursor()
                where_clause = " AND ".join(f"{key} = ?" for key in filter_dict)
                values = tuple(filter_dict.values())
                query = f"SELECT * FROM {self.table_name} WHERE {where_clause} LIMIT 1"
                cursor.execute(query, values)
                row = cursor.fetchone()
                return dict(row) if row else None

        def find(self, filter_dict: dict = None, limit: int = None, skip: int = None) -> list:
            with sqlite3.connect(database_name) as conn:
                conn.row_factory = sqlite3.Row
                cursor = conn.cursor()

                where_clause = ""
                values = []

                if filter_dict:
                    where_clause = "WHERE " + " AND ".join(f"{k} = ?" for k in filter_dict)
                    values.extend(filter_dict.values())

                limit_clause = f"LIMIT {limit}" if limit is not None else ""
                offset_clause = f"OFFSET {skip}" if skip is not None else ""

                query = f"SELECT * FROM {self.table_name} {where_clause} {limit_clause} {offset_clause}".strip()
                cursor.execute(query, tuple(values))
                results = cursor.fetchall()
                return [dict(row) for row in results]

        def update_all_rows(self, key: str, value):
            with sqlite3.connect(database_name) as conn:
                cursor = conn.cursor()
                query = f"UPDATE {self.table_name} SET {key} = ?"
                cursor.execute(query, (value,))
                conn.commit()

    class DBWrapper:
        def __getattr__(self, table_name):
            return DBFunctions(table_name)

    db = DBWrapper()

elif DB_TYPE == "mongodb":
    from motor.motor_asyncio import AsyncIOMotorClient

    DB = os.getenv("DB_NAME")
    MONGO_URL = os.getenv("MONGO_URL", "mongodb://localhost:27017")

    client = AsyncIOMotorClient(MONGO_URL)
    db = client[DB]

else:
    raise ValueError("Unsupported DB_TYPE. Must be either 'sqlite' or 'mongodb'.")



from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware


app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],  # Allow all origins
    allow_credentials=True,
    allow_methods=["*"],  # Allow all HTTP methods
    allow_headers=["*"],  # Allow all headers
)

@app.get("/")
def read_root():
    return {"message": "Hello from FasterAPI!"}

@app.get("/health")
async def health_check():
    return {"status": "healthy"}
    